import tqdm
from getpass import getpass
import json
import asyncio
import aiohttp
from itertools import islice
from typing import (
    List, TypeVar, Iterator, Sequence, AsyncIterator, Coroutine, Any
)
from collections.abc import Callable
import os
from dataclasses import dataclass
from .datasets import HubSplit, JsonlConvos, Dataset, Prompts, Convos

@dataclass
class ClientOpts:
    base_url: str
    api_key: str
    retries: int = 3

@dataclass
class Extractor:
    teacher: str
    max_concurrent: int
    output_dir: str
    client_opts: ClientOpts
    dataset: Dataset[Prompts]

    def run(self):
        coro = generate_async(self)
        try:
            loop = asyncio.get_running_loop()
            future = loop.run_corouting_threadsafe(coro)
            return future.result()
        except RuntimeError:
            # No running event loop, create one
            return asyncio.run(coro)

    def output_dataset(self) -> Dataset[Convos]:
        return Dataset(
            train=get_jsonl_convos(self.output_dir, self.dataset.train),
            eval=get_jsonl_convos(self.output_dir, (self.dataset.eval or [])),
        )

def get_jsonl_convos(output_dir: str, datasets: Sequence[Prompts]):
    output_convos: List[JsonlConvos] = []
    for prompts in datasets:
        output_convos.append(JsonlConvos(
            path=os.path.join(output_dir, prompts.output_path)
        ))
    return output_convos

async def make_request(
    session, url: str,
    headers: dict[str, str],
    payload,
    retries: int,
):
    for attempt in range(retries):
        try:
            async with session.post(url, headers=headers, json=payload) as response:
                result = await response.json()
                message = result['choices'][0]['message']
                return [
                    {
                        "role": "user",
                        "content": payload["messages"][0]["content"],
                    },
                    {
                        "role": "assistant",
                        "content": message["content"]
                    },
                ]
        except Exception as e:
            if attempt == retries - 1:
                print(f"Failed after {retries} attempts: {e}")
                raise
            await asyncio.sleep(1 * (attempt + 1))

T = TypeVar('T')
R = TypeVar('R')
async def stream_with_concurrency(
    max_concurrent: int,
    iterator: Iterator[T],
    process_fn: Callable[[T], Coroutine[Any, Any, R]],
    progress_bar: tqdm.tqdm,
) -> AsyncIterator[R]:
    """Streams items from iterator with max_concurrent in-flight operations."""
    in_flight = set()

    try:
        # Prime the pump with first max_concurrent items
        for _ in range(max_concurrent):
            item = next(iterator)
            task = asyncio.create_task(process_fn(item))
            in_flight.add(task)
            task.add_done_callback(in_flight.discard)

        # As each request completes, start a new one
        while in_flight:
            done, _ = await asyncio.wait(in_flight, return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                result = await task
                if progress_bar:
                    progress_bar.update(1)
                yield result

                try:
                    item = next(iterator)
                    task = asyncio.create_task(process_fn(item))
                    in_flight.add(task)
                    task.add_done_callback(in_flight.discard)
                except StopIteration:
                    pass

    except StopIteration:
        # Handle case where we have fewer items than max_concurrent
        if in_flight:
            for result in await asyncio.gather(*in_flight):
                if progress_bar:
                    progress_bar.update(1)
                yield result

async def process_prompts(
    max_concurrent: int,
    prompts: Prompts,
    teacher: str,
    client_opts: ClientOpts,
):
    url = f"{client_opts.base_url}/chat/completions"
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {client_opts.api_key}"
    }

    async with aiohttp.ClientSession() as session:
        with tqdm.tqdm(total=prompts.count(), desc=f"Generating {prompts.output_path} completions") as pbar:
            async def process_prompt(prompt):
                return await make_request(
                    session=session,
                    url=url,
                    headers=headers,
                    payload={
                        "model": teacher,
                        "messages": [{"role": "user", "content": prompt}],
                    },
                    retries=client_opts.retries,
                )

            async for result in stream_with_concurrency(
                max_concurrent,
                prompts.items(),
                process_prompt,
                pbar
            ):
                yield result

async def generate_for_datasets(config: Extractor, datasets: Sequence[Prompts]):
    output_convos: List[JsonlConvos] = []
    for prompts in datasets:
        output_path = os.path.join(
            config.output_dir,
            prompts.output_path,
        )
        output_convos.append(JsonlConvos(path=output_path))
        output_dir = os.path.dirname(output_path)
        os.makedirs(output_dir, exist_ok=True)

        with open(output_path, "w") as f:
            async for dialog in process_prompts(
                prompts=prompts,
                max_concurrent=config.max_concurrent,
                teacher=config.teacher,
                client_opts=config.client_opts,
            ):
                f.write(json.dumps({ "conversations": dialog }))
                f.write("\n")
    return output_convos

async def generate_async(config: Extractor):
    # Process train and test splits
    train = await generate_for_datasets(config, config.dataset.train)
    eval = await generate_for_datasets(config, (config.dataset.eval or []))

    print("Done extracting!")
    return Dataset(train=train, eval=eval)



from dataclasses import dataclass
from typing import Sequence
import unfat.axolotl as axolotl
from .datasets import Dataset, Convos

@dataclass
class LoraSettings:
    lora_r: int
    lora_alpha: int
    num_epochs: int
    lora_dropout: float = 0.01
    weight_decay: float = 0.01
    learning_rate: float = 4e-4
    warmup_steps: int = 10
    evals_per_epoch: int = 1

    def llama_70b_axolotl(self, dataset: Dataset[Convos]):
        config = axolotl.Config(
            base_model="unsloth/Meta-Llama-3.1-70B-Instruct",
            lora_r=self.lora_r,
            lora_alpha=self.lora_alpha,
            lora_dropout=self.lora_dropout,
            gradient_accumulation_steps=8,
            micro_batch_size=2,
            num_epochs=self.num_epochs,
            learning_rate=self.learning_rate,
            warmup_steps=self.warmup_steps,
            evals_per_epoch=self.evals_per_epoch,
            weight_decay=self.weight_decay,
        )
        return config.generate(dataset)



from typing import Sequence, TypedDict
from dataclasses import dataclass, field, asdict, replace
import yaml
import os
from .datasets import Dataset, Convos, HubConvos, JsonlConvos

@dataclass
class AxolotlJsonlConvos:
    path: str
    ds_type: str = "json"
    split: str = "train"
    type: str = "chat_template"

class FsdpConf(TypedDict):
    fsdp_limit_all_gathers: bool
    fsdp_sync_module_states: bool
    fsdp_offload_params: bool
    fsdp_use_orig_params: bool
    fsdp_cpu_ram_efficient_loading: bool
    fsdp_auto_wrap_policy: str
    fsdp_transformer_layer_cls_to_wrap: str
    fsdp_state_dict_type: str
    fsdp_sharding_strategy: str

def default_fsdp_conf() -> FsdpConf:
    return {
        "fsdp_limit_all_gathers": True,
        "fsdp_sync_module_states": True,
        "fsdp_offload_params": True,
        "fsdp_use_orig_params": False,
        "fsdp_cpu_ram_efficient_loading": True,
        "fsdp_auto_wrap_policy": "TRANSFORMER_BASED_WRAP",
        "fsdp_transformer_layer_cls_to_wrap": "LlamaDecoderLayer",
        "fsdp_state_dict_type": "FULL_STATE_DICT",
        "fsdp_sharding_strategy": "FULL_SHARD",
    }

@dataclass
class Config:
    base_model: str
    lora_r: int
    lora_alpha: int
    lora_dropout: float
    gradient_accumulation_steps: int
    micro_batch_size: int
    num_epochs: int
    learning_rate: float
    warmup_steps: int
    evals_per_epoch: int
    weight_decay: float

    def generate(self, dataset: Dataset[Convos]):
        eval = dataset.eval if dataset.eval else []
        return FullConfig(
            base_model=self.base_model,
            datasets=[convert_dataset(ds) for ds in dataset.train],
            test_datasets=[convert_dataset(ds) for ds in eval],
            lora_r=self.lora_r,
            lora_alpha=self.lora_alpha,
            lora_dropout=self.lora_dropout,
            gradient_accumulation_steps=self.gradient_accumulation_steps,
            micro_batch_size=self.micro_batch_size,
            num_epochs=self.num_epochs,
            learning_rate=self.learning_rate,
            warmup_steps=self.warmup_steps,
            evals_per_epoch=self.evals_per_epoch,
            weight_decay=self.weight_decay,
        )

def convert_dataset(dataset: HubConvos | JsonlConvos):
    if isinstance(dataset, HubConvos):
        return dataset
    else:
        return AxolotlJsonlConvos(path=dataset.path)

@dataclass
class FullConfig(Config):
    datasets: Sequence[AxolotlJsonlConvos | HubConvos]
    test_datasets: Sequence[AxolotlJsonlConvos | HubConvos]

    # optionally might have model_type or tokenizer_type
    model_type: str = "LlamaForCausalLM"
    tokenizer_type: str = "AutoTokenizer"

    load_in_8bit: bool = False
    load_in_4bit: bool = False
    strict: bool = False

    dataset_prepared_path: str = "last_run_prepared"

    adapter: str = "lora"

    sequence_len: int = 8192
    sample_packing: bool = False
    pad_to_sequence_len: bool = True

    lora_target_linear: bool = True

    optimizer: str = "adamw_torch_fused"
    lr_scheduler: str = "linear"

    train_on_inputs: bool = False
    group_by_length: bool = False
    bf16: str = "auto"
    tf32: bool = False

    gradient_checkpointing: bool = True
    gradient_checkpointing_kwargs: dict[str, bool] = field(default_factory=lambda: {
        "use_reentrant": True
    })
    logging_steps: int = 1
    flash_attention: bool = True

    saves_per_epoch: int = 1
    fsdp: Sequence[str] = field(default_factory=lambda: [
        "full_shard",
        "auto_wrap",
    ])
    fsdp_config: FsdpConf = field(default_factory=default_fsdp_conf)
    special_tokens: dict[str, str] = field(default_factory=lambda: {
        "pad_token": "<|end_of_text|>"
    })

    def save(self, output_dir: str):
        clone = replace(self)
        for convos in clone.datasets:
            rewrite_ds_paths(output_dir, convos)
        for convos in clone.test_datasets:
            rewrite_ds_paths(output_dir, convos)
        file_contents = yaml.dump(asdict(clone), default_flow_style=False)
        path = os.path.join(output_dir, "axolotl.yaml")
        with open(path, "w") as f:
            f.write(file_contents)

def rewrite_ds_paths(output_dir: str, convos: AxolotlJsonlConvos | HubConvos):
    if isinstance(convos, HubConvos):
        return

    convos.path = "./" + os.path.relpath(convos.path, output_dir)



from datasets import load_dataset
from dataclasses import dataclass
from typing import TypedDict, Sequence, Literal, TypeVar, Generic, Iterator
from collections.abc import Callable
import json

T = TypeVar("T")
@dataclass
class Dataset(Generic[T]):
    train: Sequence[T]
    eval: Sequence[T] | None = None

@dataclass
class HubSplit:
    name: str
    max_rows: int | None = None

@dataclass
class Prompts:
    output_path: str
    count: Callable[[], int]
    items: Callable[[], Iterator[str]]

def hub_prompts(name: str, split: str | HubSplit, text_field: str):
    split_name = split.name if isinstance(split, HubSplit) else split
    ds = load_dataset(name, split=split_name)
    max_rows = split.max_rows if isinstance(split, HubSplit) else None

    def count():
        return max_rows or len(ds)

    def items():
        count = 0
        for example in ds:
            yield example[text_field]
            count += 1
            if (max_rows is not None) and count >= max_rows:
                break

    return Prompts(
        output_path=f"hub/{name}-{split_name}.jsonl",
        count=count,
        items=items,
    )

def jsonl_prompts(path: str, name: str, text_field: str):
    def count():
        with open(path, 'r') as f:
            total_lines = sum(1 for _ in f)
    def items():
        with open(path, 'r') as f:
            for line in f:
                data = json.loads(line.strip())
                yield data[split.dataset.text_field]
    return Prompts(
        output_path=f"jsonl/{name}.jsonl",
        count=count,
        items=items,
    )

@dataclass
class JsonlConvos:
    path: str

@dataclass
class HubConvos:
    name: str
    type: str
    splits: Sequence[str | HubSplit]

Convos = JsonlConvos | HubConvos
